import urllib
import datetime
import time
import calendar
from decimal import Decimal

from django import template
from django.db.models import Sum
from django.core.urlresolvers import reverse

try:
    from django.utils import timezone
except ImportError:
    from timepiece import timezone

from dateutil.relativedelta import relativedelta
from dateutil import rrule

import timepiece.models as timepiece
from timepiece.utils import get_total_time, get_week_start, get_month_start


register = template.Library()


@register.filter
def seconds_to_hours(seconds):
    return round(seconds / 3600.0, 2)


@register.inclusion_tag('timepiece/time-sheet/bar_graph.html',
                        takes_context=True)
def bar_graph(context, name, worked, total, width=None, suffix=None):
    if not width:
        width = 400
        suffix = 'px'
    left = total - worked
    over = 0
    over_total = 0
    error = ''
    if worked < 0:
        error = 'Somehow you\'ve logged %s negative hours for %s this week.' \
        % (abs(worked), name)
    if left < 0:
        over = abs(left)
        left = 0
        total = over + total
    return {
        'name': name, 'worked': worked,
        'total': total, 'left': left,
        'over': over, 'width': width,
        'suffix': suffix, 'error': error,
        }


@register.inclusion_tag('timepiece/time-sheet/_date_filters.html',
    takes_context=True)
def date_filters(context, options=None):
    request = context['request']
    from_slug = 'from_date'
    to_slug = 'to_date'
    use_range = True
    if not options:
        options = ('months', 'quarters', 'years')

    def construct_url(from_date, to_date):
        query = request.GET.copy()
        query.pop(to_slug, None)
        query.pop(from_slug, None)
        query[to_slug] = to_date.strftime('%m/%d/%Y')
        if use_range:
            query[from_slug] = from_date.strftime('%m/%d/%Y')
        return '%s?%s' % (request.path, query.urlencode())

    filters = {}
    if 'months_no_range' in options:
        filters['Past 12 Months'] = []
        single_month = relativedelta(months=1)
        from_date = datetime.date.today().replace(day=1) + \
            relativedelta(months=1)
        for x in range(12):
            to_date = from_date
            use_range = False
            from_date = to_date - single_month
            url = construct_url(from_date, to_date - relativedelta(days=1))
            filters['Past 12 Months'].append(
                (from_date.strftime("%b '%y"), url))
        filters['Past 12 Months'].reverse()

    if 'months' in options:
        filters['Past 12 Months'] = []
        single_month = relativedelta(months=1)
        from_date = datetime.date.today().replace(day=1) + \
            relativedelta(months=1)
        for x in range(12):
            to_date = from_date
            from_date = to_date - single_month
            url = construct_url(from_date, to_date - relativedelta(days=1))
            filters['Past 12 Months'].append(
                (from_date.strftime("%b '%y"), url))
        filters['Past 12 Months'].reverse()

    if 'years' in options:
        start = datetime.date.today().year - 3

        filters['Years'] = []
        for year in range(start, start + 4):
            from_date = datetime.datetime(year, 1, 1)
            to_date = from_date + relativedelta(years=1)
            url = construct_url(from_date, to_date - relativedelta(days=1))
            filters['Years'].append((str(from_date.year), url))

    if 'quarters' in options:
        filters['Quarters (Calendar Year)'] = []
        to_date = datetime.date(datetime.date.today().year - 1, 1, 1)
        for x in range(8):
            from_date = to_date
            to_date = from_date + relativedelta(months=3)
            url = construct_url(from_date, to_date - relativedelta(days=1))
            filters['Quarters (Calendar Year)'].append(
                ('Q%s %s' % ((x % 4) + 1, from_date.year), url)
            )

    return {'filters': filters}


@register.inclusion_tag('timepiece/time-sheet/invoice/_invoice_subheader.html',
                        takes_context=True)
def invoice_subheaders(context, current):
    return {
        'current': current,
        'invoice': context['invoice'],
    }


@register.simple_tag
def hours_for_assignment(assignment, date):
    end = date + relativedelta(days=5)
    blocks = assignment.blocks.filter(
        date__gte=date, date__lte=end).select_related()
    hours = blocks.aggregate(hours=Sum('hours'))['hours']
    if not hours:
        hours = ''
    return hours


@register.simple_tag
def total_allocated(assignment):
    hours = assignment.blocks.aggregate(hours=Sum('hours'))['hours']
    if not hours:
        hours = ''
    return hours


@register.simple_tag
def hours_for_week(user, date):
    end = date + relativedelta(days=5)
    blocks = timepiece.AssignmentAllocation.objects.filter(
                                                 assignment__user=user,
                                                 date__gte=date, date__lte=end)
    hours = blocks.aggregate(hours=Sum('hours'))['hours']
    if not hours:
        hours = ''
    return hours


@register.simple_tag
def weekly_hours_worked(rp, date):
    hours = rp.hours_in_week(date)
    if not hours:
        hours = ''
    return hours


@register.simple_tag
def monthly_overtime(rp, date):
    hours = rp.total_monthly_overtime(date)
    if not hours:
        hours = ''
    return hours


@register.simple_tag
def week_start(date):
    return get_week_start(date).strftime('%m/%d/%Y')


@register.simple_tag
def get_active_hours(entry):
    """Use with active entries to obtain time worked so far"""
    if not entry.end_time:
        if entry.is_paused:
            entry.end_time = entry.pause_time
        else:
            entry.end_time = timezone.now()
    return Decimal('%.2f' % round(entry.total_hours, 2))


@register.simple_tag
def show_cal(from_date, offset=0):
    date = get_month_start(from_date)
    date = date + relativedelta(months=offset)
    html_cal = calendar.HTMLCalendar(calendar.SUNDAY)
    return html_cal.formatmonth(date.year, date.month)


@register.simple_tag
def get_uninvoiced_hours(entries):
    hours_uninvoiced = 0
    for entry in entries:
        if entry['status'] != 'invoiced' and entry['status'] != 'not-invoiced':
            hours_uninvoiced += entry['s']
    return hours_uninvoiced


@register.filter
def work_days(end):
    weekdays = (rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR)
    days = rrule.rrule(rrule.DAILY, byweekday=weekdays,
                       dtstart=datetime.date.today(), until=end)
    return len(list(days))


@register.simple_tag
def timesheet_url(type, pk, date):
    if type == 'project':
        name = 'project_time_sheet'
    elif type == 'user':
        name = 'view_person_time_sheet'

    url = reverse(name, args=(pk,))
    params = {
        'month': date.month,
        'year': date.year
    }

    return '?'.join((url, urllib.urlencode(params),))



import datetime
from decimal import Decimal

from django.core.urlresolvers import reverse
from django.contrib.auth.models import User

try:
    from django.utils import timezone
except ImportError:
    from timepiece import timezone

from timepiece.templatetags.timepiece_tags import get_active_hours
from timepiece.tests.base import TimepieceDataTestCase
from timepiece import utils


class DashboardTestCase(TimepieceDataTestCase):

    def setUp(self):
        super(DashboardTestCase, self).setUp()
        self.unpriveleged_user = User.objects.create_user(
            username='tester',
            password='abc',
            email='email@email.com'
        )
        self.url = reverse('timepiece-entries')
        self.text = [u'Clock In', u'Add Entry', u'My Active Entries']
        self.now = timezone.now()

    def test_unpriveleged_user(self):
        """
        A regular user should not be able to see what people are
        working on or timesheet related links
        """
        self.client.login(username='tester', password='abc')

        response = self.client.get(self.url)
        for text in self.text:
            self.assertNotContains(response, text)

    def test_timepiece_user(self):
        """
        A timepiece user should be able to see what others are
        working on as well as timesheet links
        """
        self.client.login(username='user', password='abc')

        response = self.client.get(self.url)
        for text in self.text:
            self.assertContains(response, text)

    def test_work_this_week(self):
        """
        Work This Week table should contain hours worked on projects this week,
        including from active entries.
        """
        projects = []
        projects.append(self.create_project(billable=True))
        projects.append(self.create_project(billable=False))
        past = []
        past.append(self.create_entry({  # 15 minutes
            'start_time': self.now + datetime.timedelta(minutes=5),
            'end_time': self.now + datetime.timedelta(minutes=20),
            'project': projects[0],
            'activity': self.create_activity(data={'billable': True}),
        }))
        past.append(self.create_entry({  # 15 minutes
            'start_time': self.now + datetime.timedelta(minutes=25),
            'end_time': self.now + datetime.timedelta(minutes=40),
            'project': projects[0],
            'activity': self.create_activity(data={'billable': True}),
        }))
        past.append(self.create_entry({  # 60 minutes
            'start_time': self.now + datetime.timedelta(minutes=45),
            'end_time': self.now + datetime.timedelta(minutes=105),
            'project': projects[1],
            'activity': self.create_activity(data={'billable': False}),
        }))
        current = self.create_entry({  # 30 minutes
            'start_time': self.now - datetime.timedelta(minutes=30),
            'project': projects[0],
            'activity': self.create_activity(data={'billable': True}),
        })
        current_hours = get_active_hours(current)
        total_hours = sum([p.hours for p in past]) + current_hours

        self.client.login(username='user', password='abc')
        response = self.client.get(self.url)
        context = response.context
        self.assertEquals(len(context['my_active_entries']), 1)
        self.assertEquals(get_active_hours(context['my_active_entries'][0]),
                current_hours)
        self.assertEquals(len(context['project_entries']), 2)
        for entry in context['project_entries']:
            if entry['project__pk'] == projects[0].pk:
                self.assertEquals(entry['sum'], Decimal('0.50'))
            else:
                self.assertEquals(entry['sum'], Decimal('1.00'))
        self.assertEquals(len(context['activity_entries']), 2)
        for entry in context['activity_entries']:
            self.assertEquals(entry['sum'], Decimal('1.00'))
        self.assertEquals(context['current_total'], total_hours)

    def test_time_detail(self):
        """
        An active entry should not appear in the Time
        Detail This Week table
        """
        entry = self.create_entry({
            'start_time': self.now,
            'end_time': self.now + datetime.timedelta(hours=1),
        })
        self.create_entry({
            'start_time': self.now + datetime.timedelta(hours=2),
            'project': self.create_project(billable=True)
        })

        self.client.login(username='user', password='abc')

        response = self.client.get(self.url)
        context = response.context
        self.assertEquals(len(context['this_weeks_entries']), 1)
        self.assertEquals(context['this_weeks_entries'][0], entry)



from django.conf import settings

from timepiece import models as timepiece
from timepiece.forms import QuickSearchForm


def timepiece_settings(request):
    default_famfamfam_url = settings.STATIC_URL + 'images/icons/'
    famfamfam_url = getattr(settings, 'FAMFAMFAM_URL', default_famfamfam_url)
    context = {
        'FAMFAMFAM_URL': famfamfam_url,
    }
    return context


def quick_search(request):
    return {
        'quick_search_form': QuickSearchForm(),
    }


def active_entries(request):
    active_entries = None

    if request.user.is_authenticated():
        active_entries = timepiece.Entry.objects.filter(
            end_time__isnull=True,
        ).exclude(
            user=request.user,
        ).select_related('user', 'project', 'activity')

    return {
        'active_entries': active_entries,
    }


def extra_nav(request):
    context = {
        'extra_nav': getattr(settings, 'EXTRA_NAV', {})
    }
    return context



import datetime
import random
from decimal import Decimal

from django.conf import settings
from django.core.urlresolvers import reverse
from django.db.models import Sum
from django.contrib.auth.models import User

try:
    from django.utils import timezone
except ImportError:
    from timepiece import timezone

from timepiece import models as timepiece
from timepiece import forms as timepiece_forms
from timepiece.tests.base import TimepieceDataTestCase

from dateutil import relativedelta

from timepiece.projection import run_projection, user_weekly_assignments
from timepiece import utils


class ProjectionTest(TimepieceDataTestCase):

    def setUp(self):
        person = User.objects.create_user('test', 'a@b.com', 'abc')
        self.ps = self.create_person_schedule(data={'user': person})

    def log_time(self, assignment, delta=None, start=None):
        if delta:
            hours, minutes = delta
        else:
            hours = 4
            minutes = 0
        if not start:
            start = timezone.now()
        elif not isinstance(start, datetime.datetime):
            start = datetime.datetime.combine(start, datetime.time())
        end = start + datetime.timedelta(hours=hours, minutes=minutes)
        data = {'user': assignment.user,
                'start_time': start,
                'end_time': end,
                'project': assignment.contract.project}
        return self.create_entry(data)

    def test_week_start(self):
        """ Test that all days Sun. through Sat. return the previous Monday"""
        monday = datetime.date(2011, 1, 10)
        self.assertEqual(monday, utils.get_week_start(monday).date())
        sunday = datetime.date(2011, 1, 16)
        self.assertEqual(monday, utils.get_week_start(sunday).date())
        following_monday = datetime.date(2011, 1, 17)
        saturday = datetime.date(2011, 1, 22)
        self.assertEqual(following_monday,
            utils.get_week_start(saturday).date())

    def  test_month_start(self):
        """ Test that any day returns the first day of the month"""
        days = [
            datetime.date(2011, 1, 1),
            datetime.date(2011, 1, 16),
            datetime.date(2011, 1, 17),
            datetime.date(2011, 1, 22),
            datetime.date(2011, 1, 31),
        ]
        date = utils.add_timezone(datetime.datetime(2011, 1, 1))
        for day in days:
            self.assertEqual(utils.get_month_start(day), date)

    def test_generate_dates(self):
        """ Test generation of full date ranges """
        ### test WEEKLY
        # 2 weeks
        start = datetime.date(2011, 1, 17)
        end = datetime.date(2011, 1, 29)
        weeks = utils.generate_dates(start=start, end=end)
        self.assertEqual(2, weeks.count())
        # 3 weeks
        start = datetime.date(2011, 1, 17)
        end = datetime.date(2011, 1, 31)
        weeks = utils.generate_dates(start=start, end=end)
        self.assertEqual(3, weeks.count())
        # random weeks
        num = random.randint(5, 20)
        start = utils.get_week_start(datetime.date.today())
        end = start + datetime.timedelta(weeks=num - 1)
        weeks = utils.generate_dates(start=start, end=end)
        self.assertEqual(num, weeks.count())
        ### test MONTHLY
        start = datetime.date(2011, 1, 17)
        end = datetime.date(2011, 4, 29)
        months = utils.generate_dates(start=start, end=end, by='month')
        self.assertEqual(4, months.count())
        for index, month in enumerate(months):
            self.assertEqual(month.date(), datetime.date(2011, index + 1, 1))
        ### test DAILY
        start = datetime.date(2011, 2, 1)
        end = datetime.date(2011, 2, 15)
        days = utils.generate_dates(start=start, end=end, by='day')
        self.assertEqual(15, days.count())
        for index, day in enumerate(days):
            self.assertEqual(day.date(), datetime.date(2011, 2, index + 1))

    def test_week_window(self):
        """ Test generation of weekly window with given date """
        #Tuesday
        day = datetime.date(2011, 2, 1)
        expected_start = datetime.date(2011, 1, 31)
        expected_end = datetime.date(2011, 2, 7)
        start, end = utils.get_week_window(day)
        self.assertEqual(start.toordinal(), expected_start.toordinal())
        self.assertEqual(end.toordinal(), expected_end.toordinal())

    def test_project_contract_remaining_weeks(self):
        """ Test calculation of contract remaining weeks """
        num = random.randint(5, 20)
        start = datetime.date.today()
        end = start + datetime.timedelta(weeks=num - 1)
        pc = self.create_project_contract({'num_hours': 100, 'end_date': end,
                                           'start_date': start})
        self.assertEqual(num, pc.weeks_remaining.count())

    def test_weekly_commmitment(self):
        """ Test calculation of contract assignment's weekly commitment """
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=40)
        self.assertEqual(ca.weekly_commitment(start), 40)

    def test_weekly_commmitment_with_earlier_allocation(self):
        """ Test calculation of contract assignment's weekly commitment """
        # 1 week assignment, 20 hours
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=1) - datetime.timedelta(days=1)
        ca1 = self._assign(start, end, hours=20)
        # allocate 20 hours to this assignment
        ca1.blocks.create(date=start, hours=20)
        # 1 week assignment, 20 hours
        ca2 = self._assign(start, end, hours=20)
        # only 20 hours left this week
        self.assertEqual(ca2.weekly_commitment(start), 20)

    def test_weekly_commmitment_with_look_ahead(self):
        """ Test later assignment's min hours factor into weekly commitment """
        # 1 week assignment, 40 hours
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=1) - datetime.timedelta(days=1)
        ca1 = self._assign(start, end, hours=40)
        # 2 week assignment, 40 hours
        end = start + datetime.timedelta(weeks=1) - datetime.timedelta(days=1)
        ca2 = self._assign(start, end, hours=40)
        ca2.min_hours_per_week = 5
        ca2.save()
        self.assertEqual(ca1.weekly_commitment(start), 35)
        ca1.blocks.create(date=start, hours=35)
        self.assertEqual(ca2.weekly_commitment(start), 5)

    def test_weekly_commmitment_with_hours_worked(self):
        """ Test weekly commitment with previously logged hours """
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=30)
        self.log_time(ca, start=start, delta=(10, 0))
        self.assertEqual(ca.hours_worked, 10)
        self.assertEqual(ca.hours_remaining, 20)
        self.assertEqual(ca.weekly_commitment(start), 30)

    def test_weekly_commitment_over_remaining(self):
        # 1 week assignment, 20 hours
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=1) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=20)
        # only 20 hours left on assignment
        self.assertEqual(ca.weekly_commitment(start), 20)

    def _assign(self, start=None, end=None, hours=30):
        pc = self.create_project_contract({'num_hours': hours,
                                           'start_date': start,
                                           'end_date': end})
        ca = self.create_contract_assignment({'contract': pc,
                                              'user': self.ps.user,
                                              'num_hours': hours})
        return ca

    def test_assignment_active_ends_mid_week(self):
        """ Test manager returns assignments that end before end of window """
        start = utils.get_week_start() - datetime.timedelta(days=2)
        end = start + datetime.timedelta(weeks=1)
        ca = self._assign(start, end)
        week = utils.get_week_start()
        next_week = week + datetime.timedelta(weeks=1)
        assignments = timepiece.ContractAssignment.objects
        assignments = assignments.active_during_week(week, next_week)
        self.assertTrue(assignments.filter(pk=ca.pk).exists())

    def test_assignment_active_starts_mid_week(self):
        """ Test manager returns assignments that start before window """
        start = utils.get_week_start() + datetime.timedelta(days=2)
        end = start + datetime.timedelta(weeks=2)
        ca = self._assign(start, end)
        week = utils.get_week_start()
        next_week = week + datetime.timedelta(weeks=1)
        assignments = timepiece.ContractAssignment.objects
        assignments = assignments.active_during_week(week, next_week)
        self.assertTrue(assignments.filter(pk=ca.pk).exists())

    def test_assignment_active_within_week(self):
        """ Test manager returns assignments that contain entire week """
        start = utils.get_week_start() - datetime.timedelta(weeks=1)
        end = start + datetime.timedelta(weeks=3)
        ca = self._assign(start, end)
        week = utils.get_week_start()
        next_week = week + datetime.timedelta(weeks=1)
        assignments = timepiece.ContractAssignment.objects
        assignments = assignments.active_during_week(week, next_week)
        self.assertTrue(assignments.filter(pk=ca.pk).exists())

    def test_no_remaining_hours(self):
        """ Gurantee no overcommittment """
        # 1 week, 40 hours
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=1) - datetime.timedelta(days=1)
        ca1 = self._assign(start, end, hours=40)
        ca1.blocks.create(date=start, hours=40)
        self.assertEqual(ca1.weekly_commitment(start), 0)
        # 2 weeks, 40 hours
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca2 = self._assign(start, end, hours=40)
        self.assertEqual(ca2.weekly_commitment(start), 0)

    def test_single_assignment_projection(self):
        # 2 weeks, 60 hours
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=60)
        run_projection()
        self.assertEqual(60, ca.blocks.aggregate(s=Sum('hours'))['s'])

    def test_min_hours_per_week_weighted(self):
        """
        Test minimum hours/week with weighting based on assignment end date
        """
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=1)
        ca1 = self._assign(start, end, hours=40)
        ca2 = self._assign(start, end + datetime.timedelta(days=1), hours=40)
        ca1.min_hours_per_week = 30
        ca1.save()
        ca2.min_hours_per_week = 30
        ca2.save()
        run_projection()
        projection = ca1.blocks.filter(date=start).aggregate(s=Sum('hours'))
        self.assertEqual(30, projection['s'])
        projection = ca2.blocks.filter(date=start).aggregate(s=Sum('hours'))
        self.assertEqual(10, projection['s'])

    def test_unallocated_hours(self):
        """ Test unallocated hours calculation """
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=40)
        unallocated_hours = ca.unallocated_hours_for_week(start)
        self.assertEqual(unallocated_hours, 40)
        ca.blocks.create(date=start, hours=5)
        unallocated_hours = ca.unallocated_hours_for_week(start)
        self.assertEqual(unallocated_hours, 35)

    def test_this_weeks_priority_type(self):
        """ Test categories for this week. """
        start = utils.get_week_start(datetime.date.today())
        end = start + datetime.timedelta(weeks=1)
        ca_starting = self._assign(start, end, hours=40)
        self.assertEqual('starting', ca_starting.this_weeks_priority_type)
        end = utils.get_week_start()
        start = end - datetime.timedelta(days=2)
        ca_ending = self._assign(start, end, hours=40)
        self.assertEqual('ending', ca_ending.this_weeks_priority_type)
        start = utils.get_week_start()
        end = start + datetime.timedelta(days=6)
        ca_starting_ending = self._assign(start, end, hours=40)
        self.assertEqual('ending', ca_starting_ending.this_weeks_priority_type)
        start = utils.get_week_start() - datetime.timedelta(days=1)
        end = start + datetime.timedelta(days=9)
        ca_ongoing = self._assign(start, end, hours=40)
        self.assertEqual('ongoing', ca_ongoing.this_weeks_priority_type)
        ## Need to test order goes ending, starting, ongoing.
        assignments = timepiece.ContractAssignment.objects.sort_by_priority()

    def test_this_weeks_allocations(self):
        # 2 weeks, 60 hours
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=20)
        person = User.objects.create_user('test2', 'a@b.com', 'abc')
        ps = self.create_person_schedule(data={'user': person})
        run_projection()
        assignments = timepiece.AssignmentAllocation.objects.during_this_week(
            self.ps.user)
        self.assertEquals(assignments.count(), 1)
        assignments = timepiece.AssignmentAllocation.objects.during_this_week(
            person)
        self.assertEquals(assignments.count(), 0)
        ca_2 = self._assign(start, end, hours=30)
        run_projection()
        assignments = timepiece.AssignmentAllocation.objects.during_this_week(
            self.ps.user)
        self.assertEquals(assignments.count(), 2)

    def test_this_weeks_hours(self):
        start = utils.get_week_start()
        end = start + datetime.timedelta(weeks=2) - datetime.timedelta(days=1)
        ca = self._assign(start, end, hours=60)
        run_projection()
        self.log_time(ca, start=start, delta=(10, 0))
        assignments = timepiece.AssignmentAllocation.objects.during_this_week(
            self.ps.user)
        self.assertEquals(assignments.count(), 1)
        assignment = assignments[0]
        self.assertEquals(assignment.hours_worked, 10)
        self.assertEquals(assignment.hours_left, assignment.hours - 10)
